/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1998
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *       Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: iostream.h,v 1.6 1999/01/09 18:36:16 jordan Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *   This module sets a generic layer on top of the standard io functions
 *   like send, recv, read and write.  It replaces the channel numbers by
 *   a generic descriptor and calls the approriate io functions.
 */


#ifndef __IOSTREAM_H__
#define __IOSTREAM_H__

/* control requests to be passed to the io_ctrl () function defined, below. As a
   general rule, a successful request returns a non-negative value. Upon error,
   -1 is returned while the global variable errno is set to some value indicating
   the error type. */

typedef enum {

  /* The sender defines a new size of the maximal data block that can be passed 
     at once.  The request returns the current buffer size (which might be 
     different from the argument size due so some adjustments,) or -1 to indicate 
     some error. */

  IO_RESIZE_BUF,		/* arg is (int*)buffer_size, or 0 */
  
  /* The sender defines a new session key which is generated internally from a 
     random source.  If no argument is given, the session key is changed immediately.  
     An integer value passed by the argument pointer defines a key schedule: A
     positive value sets the time interval to the next change while a negative
     value implies a time period, when the key is changed at regular intervals.
     A value 0 disables any key change schedule.  The request returns a non negative
     value indicating the old setting of the time due, or the key schedule. Upon 
     error -1 is returned. */
  
  IO_CHANGE_KEY,		/* arg is (int*)key, or 0 */

  /* A thread is an independent instance of the sender descriptor piping through 
     the same io channel. There is always just one active thread. The thread ID is 
     the return code passedd back from the registry service (unless -1 to indicate
     an error.)  */

  IO_REGISTER_THREAD,		/* prepare threaded channel, arg is ignored */
  IO_ACTIVATE_THREAD,		/* open channel, arg is (int*)threadID or 0 */
  
  /* Before forking, the sender has a new thread registered on sender and receiver.
     After forking, the child activates the thread while the parent unlinks the
     thread without notifying the receiver.  When the child finishes, it destroys
     the tread, finally. Upon success, the destrox/unlink functions return the id 
     of the affected thread, and -1, otherwise while errno is set to indicate the
     type of error.

     Example:

     int id = io_ctrl (socket, IO_REGISTER_THREAD, 0, 1);
     ...
     if (fork () == 0) {
       // child
       io_ctrl (socket, IO_ACTIVATE_THREAD, &id, 1);
       ...
       io_close(socket); // implies io_ctrl (socket, IO_DESTROY_THREAD, &id, 1);
       exit (0);
     }
     // parent
     io_ctrl (socket, IO_UNLINK_THREAD, &id, 1);
     ...
  */

  IO_UNLINK_THREAD,		/* close thread on sender, arg is (int*)threadID or 0 */
  IO_DESTROY_THREAD,		/* close threaded channel, arg is (int*)threadID or 0 */

  /* The number of active threads can be controlled by the following request. Note, 
     that receiver will be most likely unable to resyncronize, when this limit is
     exeeded. All that can be done is to close, and reopen the dialogue. */

  IO_MAX_THREADS,		/* set max num of threads, arg is (int*)MAX, or 0 */

  /* With the following commands, a stop-on-empty-frame is controlled. At some
     requests, the sender needs to inform the receiver about internal changes of the
     io channel state. This is done by sending an empty frame with embedded commands. 
     In IO_STOPONEMPTY_STATE with an argumet pionter to a zero value, the application
     does not see that empty frame (which is the default). If a non-zero argument is
     pointed, to these empty frames are returned to the io_recv () calling process.
     The io_ctrl function returns the state of the stop-on-empty-frame flag that was 
     active before the change took place. */

  IO_STOPONEMPTY_STATE,		/* set stop-on-empty flag, arg is (int*)BOOL, or 0 */
  
  /* When an empty frame is received, the eof flag will be set.  This can be used
     to distinguish a real empty frame in IO_STOPONEMPTY_STATE active from an
     empty frame caused by an embedded command, in which case the eof-state is
     false.  In order to read on after having received an empty frame, the eof
     state must be reset.  The io_ctrl function return the state of the eof flag that 
     was active before the change took place. */

  IO_EOF_STATE,			/* set eof flag, arg is (int*)BOOL, or 0 */

  /* The underlying protocol layer updates two counters, keeping track of the user bytes
     processed and the total protocol overhead (including the user payload.) The latter
     counter does not take into account the protocol overhead that is induced by using
     the TCP/IP (or whatever) transport protocol. Note, that the interface to reading out
     the counters is made up by an int data type, which may overflow causeing a negative
     return value.  In this case, the errno register is set to 0. */

  IO_TOTAL_COUNTER,		/* get/set total protocol data bytes, arg is (int*) or 0 */
  IO_PAYLOAD_COUNTER,		/* get/set user payload data bytes, arg is (int*) or 0 */

} io_ctrl_request ;

/* prototype of a high level io function */
typedef int (*io_fn) (void *desc, const void *msg, unsigned len, unsigned flags) ;

/* prototype of a generic ioctl function */
typedef int (*ctl_fn) (void *desc, io_ctrl_request, int *arg) ;

/* prototype of a destuctor function */
typedef void (*destroy_fn) (void *desc, unsigned fd) ;

/* layer create/destroy management */
#define      IO_PUSH_EXCLUSIVELY 4
extern void *io_push (unsigned fd, unsigned size, io_fn, ctl_fn, 
		      void (*flush)(void*), 
		      destroy_fn, unsigned how);	/* never NULL */
extern void  io_pop (unsigned fd, unsigned how) ;

/* some generic control function passing, the last argument is either 0, or
   1, where 0 == receiver, 1 == sender. */
extern int io_ctrl (unsigned, io_ctrl_request, int *, unsigned) ;


/* start/stop (default == start) management, the last argument is either 0, 1, or
   2 where 0 == receiver, 1 == sender, 2 == both. */
extern void  io_enable (unsigned fd, unsigned how) ;
extern void io_suspend (unsigned fd, unsigned how) ;


/* io top layer functions, emulates send/recv/shutdown */
extern int io_send (unsigned fd, const char *buf, unsigned len, unsigned flags);
extern int io_recv (unsigned fd,       char *buf, unsigned len, unsigned flags);
extern void rd_flush (unsigned fd) ;
extern int io_shutdown (unsigned fd, unsigned how);
extern int io_close (unsigned fd);


/* replacement for network read/write functions */
#ifndef __IOSTREAM_PRIVATE__
#define send(fd, buf, n, flg)   io_send (fd, buf, n, flg)
#define recv(fd, buf, n, flg)   io_recv (fd, buf, n, flg)
#define shutdown(fd,flg)        io_shutdown (fd, flg)
#define close(fd)               io_close (fd)
#endif

#endif /* __IOSTREAM_H__ */
