/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1998
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *   Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: peks-client.c,v 1.6 1999/01/09 18:36:12 jordan Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   PEKS - private exponent key stuff for Diffie Hellman and El Gamal
 */

#include "common-stuff.h"

#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#else
# include <socket.h> /* ????? */
#endif

#include "peks-internal.h"
#include "crandom.h"
#include "messages.h"
#include "cbc-frame.h"
#include "baseXX.h"

#define __IOSTREAM_PRIVATE__ /* prevent from using send/recv auto-replaces */
#include "iostream.h"

/* ---------------------------------------------------------------------------- *
 *                  public key handshake functions                              *
 * ---------------------------------------------------------------------------- */

static peks_key *
peks_client_setup
  (const char   *in,
   const char *host,
   const char *file)
{
  peks_key *key ;
  
  if ((key = accept_public_key_str (in)) == 0) {
    errno = errnum (ACCEPT_PUBKEY_ERR) ;
    return 0;
  }
  
  /* check whether we know the server key, already */
  if (check_peks_server_key (key, host, file) < 0) {
    end_peks_key (key) ;
    return 0 ;
  }

  /* update the key structure for sending */
  update_accepted_public_key_str (in, key) ;
  
  return key ;
}


static char *
peks_client_encode_receiver_key
  (peks_key      *k,
   const char    *s,
   unsigned     len,
   const char *type)
{
  char *t, *sk, *sl ;

  if (len == 0)
    len = strlen (s) ;
  if (type == 0)
    type = "" ;

  /*
   * make the public peks encrypted receiver session key as
   *
   *	s [] := ("key:" <base64-key-bits> <cipher-type>)
   */
  sk = bin2base64 (s, len) ;
  sl  = ALLOCA (8 + strlen (sk) + strlen (type));
  sprintf (sl, "key: %s %s\n", sk, type) ;
  XFREE (sk) ;

  t = make_elg_response_key_str (k, sl, strlen (sl), 0) ;
  DEALLOCA (sl) ;

  return t ;
}


static char *
peks_client_unwrap_sender_key
  (unsigned    *N,
   const char **T,
   const char *in)
{
  char *s, *sk, *sl;
  unsigned dummy ;

  if (N == 0)
    N = &dummy ;

  /* must be editbale in order to use strtok () */
  s = ALLOCASTRDUP (in) ;

  /* parse the sender session key from the server:
   *
   *	key: <base64-key-bits> <cipher-type>
   */

  if ((sk = strtok (s, PEKS_FS)) == 0 ||
      strcmp (sk, "key:")         != 0 ||
      (sk = strtok  (0, PEKS_FS)) == 0 ||
      (sl = strtok  (0, PEKS_FS)) == 0) {
    DEALLOCA (s);
    errno = errnum (DECODE_SRVKEY_ERR) ;
    return 0 ;
  }

  /* extract the binary key */
  sk = base64toBin (sk, N) ;

  /* dup the cypher type */
  if (T != 0)
    *T = XSTRDUP (sl) ;

  /* clean up and return */
  DEALLOCA (s);
  return sk;
}

/* ---------------------------------------------------------------------------- *
 *                  public key client handshake function                        *
 * ---------------------------------------------------------------------------- */

int
client_negotiate_session_key
  (const char  * cipher,
   int           socket,
   const char  * server,
   const char * keyfile)
{
  peks_key *key ;
  char *s, *u, *buf ;
  int n, l ;

  initRandomGenerator (0);

  /* read the public key from the server host */
  buf = ALLOCA (CBC_IO_BUFLEN_MAX+1) ;
  if ((n = recv (socket, buf, CBC_IO_BUFLEN_MAX, 0)) < 0) {
    DEALLOCA (buf) ;
    return -1 ;
  }
  buf [n] = 0 ;

  if ((key = peks_client_setup (buf, server, keyfile)) == 0) {
    DEALLOCA (buf) ;
    return -1 ;
  }

  /* generate the receiver session key s of length l */
  s = ALLOCA (l = cbc_keylen (cipher)) ;
  setRandomBytes (s, l) ;
  
  /* install the receiver session key, here */
  if ((n = cbc_recv_init (socket, cipher, s, l)) != 0) {
    end_peks_key (key) ;
    DEALLOCA (s);
    DEALLOCA (buf) ;
    errno = errnum (CLNT_INSTALL_SKEY) ;
    return -1;
  }
 
  /*  wrap the client session key and send it */
  u = peks_client_encode_receiver_key (key, s, l, cipher) ;
  end_peks_key (key) ;			/* not needed, any longer */
  DEALLOCA (s);
  if (u == 0) {
    DEALLOCA (buf) ;
    return -1 ;
  }
  send (socket, u, strlen (u), 0);
  XFREE (u) ;
  
  /* receive the sender session key from the server */
  if ((n = io_recv (socket, buf, CBC_IO_BUFLEN, 0)) < 0) {
    DEALLOCA (buf) ;
    return -1 ;
  }
  buf [n] = '\0' ;

  s = peks_client_unwrap_sender_key (&l, (const char **)&u, buf) ;
  DEALLOCA (buf) ;			/* not needed, any longer */
  if (s == 0)
    return -1 ;
    
  /* install the sender session key s of length l and type t */
  n = cbc_send_init (socket, cipher, s, l) ;
  DEALLOCA (s) ;
  if (n != 0) {
    errno = errnum (CLNT_INSTALL_CKEY) ;
    return -1 ;
  }
 
  if (io_send (socket, PEKS_MAGIC, strlen (PEKS_MAGIC)+1, 0) !=
      strlen (PEKS_MAGIC)+1) {
    errno = errnum (CLNT_MAGIC_FAILED) ;
    return -1;
  }
  
#if 0
  /*********************************************/
  buf = ALLOCA (1024) ;
  n = io_ctrl (socket, IO_CHANGE_KEY, 0, 1);
  n = io_recv (socket, buf, 1024, 0) ;
  n = io_ctrl (socket, IO_CHANGE_KEY, 0, 1);
  n = io_recv (socket, buf, 1024, 0) ;
  n = io_ctrl (socket, IO_CHANGE_KEY, 0, 1);
  n = io_recv (socket, buf, 1024, 0) ;
  n = io_ctrl (socket, IO_CHANGE_KEY, 0, 1);
  n = io_recv (socket, buf, 1024, 0) ;
  /*********************************************/
#endif

  return 0;
}
